#include	<stdio.h>
#include	<math.h>
#include	"gene.h"

int	PopulationSize = 36;
int	GenerationLength =
#ifndef NOMETABOLISM
400;
#else
50;
#endif
int	SeedForRandom = 0;
int	MaxCells = 4096;
int	IScale = 100;
char	FullScreen = 0, BlackBackground = 0;
SelectionMode	SelMode = SelGlobal3rd; /*, SelLocal3rd or SelLocalSel */
CrossoverMode	CrossMode = CrsOnePoint; /*, CrsTwoPoints or CrsUniform */
double	MutationRate = 0.01;

double	CellSize = .1;
unsigned long	PopSize, Generation, Step, BestScore, AverageScore;
Individual	*Population;
Individual	*MonitoredInd = NULL;
extern	long	DrawWhen;

main(argc, argv) int argc; char *argv[];
{
	gr_init(argc,argv);
	mem_init();
	world_reset();
	data_init(argc,argv);
	if (!DrawWhen) redraw_world();
	event_loop();
}

Gene	*random_gene()
{
	static	Gene	gene;
	unsigned long	r = lrand48();
	gene.state = r & 0xff;
	gene.alpha = (r >> 8) & 0xff;
	gene.beta = (r >> 16) & 0xff;
	gene.gamma = (r >> 24) & 0xff;
	return 	&gene;
}
world_reset()
{
	int	i, j;
	for (i = 0; i < PopulationSize; i ++) for (j = 0; j < GeneLen; j ++)
		Population[i].gene[j] = *random_gene();
	Generation = 0;
	viewe_reset();
	new_generation();
}

new_generation()
{
	int	i;
	Individual	*ind;
	static double	ax[4][4] = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
	mem_reset();
	PopSize = Step = 0;
	for (i = 0, ind = Population; i < PopulationSize; i++, ind++) {
		ind->score = ind->size = 0;
		new_cell(ind,ax,0,NULL);
	}
	Generation ++;
	if (DrawWhen) redraw_world();
	draw_info();
	if (MonitoredInd) viewg(MonitoredInd);
}

#ifndef NOMETABOLISM
#define	fitness(ind)	(ind->size)
#else
int fitness(ind) Individual *ind;
{
	Cell	*c;
	int	n = 0;
	double	dmax = 0, d;
	for (c = ind->cells; c; c = c->next) {
		if ((c->state & 0xf8) == 0) return 0;
		else n ++;
		d = c->x*c->x + c->y*c->y + c->z*c->z;
		if (d > dmax) dmax = d;
	}
	return (int)
	(n / (0.1 * dmax + 0.2 * sqrt(dmax) + 0.7 * pow(dmax,0.25)));
}
#endif

exec_one_step()
{
	int	i, drawflag;
	Individual	*ind;
	Cell	*c;
	Step ++;
	move_view_point();
	for (i = 0, ind = Population; i < PopulationSize; i ++, ind ++)
	if (ind->size < MaxCells) {
		for (c = ind->cells; c; c = c->next) cell_inc_nut(c);
		for (c = ind->cells; c; c = c->next) cell_move_nut(c);
		for (c = ind->cells; c; c = c->next) cell_step(ind,c);
	}
	drawflag = DrawWhen? (Step % DrawWhen == 0)
		: (Step == GenerationLength);
	if (drawflag) redraw_world();
	if (Step == GenerationLength) {
		BestScore = AverageScore = 0;
		for (i = 0, ind = Population; i < PopulationSize; i++, ind++) {
			ind->score = fitness(ind);
			if (ind->score > BestScore) BestScore = ind->score;
			AverageScore += ind->score;
		}
		AverageScore /= PopulationSize;
		viewe_display(BestScore,AverageScore);
	}
	if (drawflag) draw_info();
	return (Step < GenerationLength);
}

static	Gene	*GenePool = NULL;
static int work_comp(a,b) Individual *a, *b;
{
	return (a->score > b->score)? -1 : (a->score < b->score);
}
selection_G3()
{
	int	i, j, k, mask, m;
	Individual	*p, *c;
	qsort((char *)Population,PopulationSize,sizeof(Individual),work_comp);
	k = PopulationSize/3;
	for (i = 0, p = Population, c = Population + k; i < k; i++, p++, c++)
		crossover(c,p,c->gene);
	m = PopulationSize - k * 2;
	for (i = 0; i < m; i ++)
		mutation(Population[i].gene,Population[k*2+i].gene);
	new_generation();
}
selection_L3()
{
	int	i, j, k, mask, m;
	Individual	*p, *q, *r, *best;
	Gene	*gp;
	if (!GenePool) {
		GenePool = (Gene *)malloc(sizeof(Gene)*GeneLen*PopulationSize);
		if (!GenePool) { perror("malloc for selection"); exit(1); }
	}
	for (i = 0; i < PopulationSize; i ++) {
		p = Population + i;
		q = Population + (i - 1 + PopulationSize) % PopulationSize;
		r = Population + (i + 1) % PopulationSize;
		gp = GenePool + GeneLen * i;
		if (p->score > q->score && p->score > r->score)
			memcpy(gp,p->gene,sizeof(Gene)*GeneLen);
		else {
			best = (q->score > r->score)? q : r;
			if (p->score <= q->score && p->score <= r->score)
				mutation(best->gene,gp);
			else crossover(p,best,gp);
		}
	}
	for (i = 0; i < PopulationSize; i ++)
		memcpy(Population[i].gene,GenePool+GeneLen*i,
			sizeof(Gene)*GeneLen);
	new_generation();
}
selection_LS()
{
	int	i, j, k, mask, m;
	Individual	*p, *q, *r, *p1, *p2;
	Gene	*gp;
	if (!GenePool) {
		GenePool = (Gene *)malloc(sizeof(Gene)*GeneLen*PopulationSize);
		if (!GenePool) { perror("malloc for selection"); exit(1); }
	}
	for (i = 0; i < PopulationSize; i ++) {
		p = Population + i;
		q = Population + (i - 1 + PopulationSize) % PopulationSize;
		r = Population + (i + 1) % PopulationSize;
		gp = GenePool + GeneLen * i;
		if (p->score > q->score) {
			p1 = p;
			p2 = (q->score > r->score)? q : r;
		} else {
			p1 = q;
			p2 = (p->score > r->score)? p : r;
		}
		crossover(p1,p2,gp);
		mutation(gp,gp);
	}
	for (i = 0; i < PopulationSize; i ++)
		memcpy(Population[i].gene,GenePool+GeneLen*i,
			sizeof(Gene)*GeneLen);
	new_generation();
}
selection()
{
	switch (SelMode) {
		case SelGlobal3rd : selection_G3(); break;
		case SelLocal3rd : selection_L3(); break;
		case SelLocalSel : selection_LS();
	}
}

crossover(p1,p2,gd)
Individual	*p1, *p2;
Gene	*gd;
{
	char	*cp1 = (char *)(p1->gene), *cp2 = (char *)(p2->gene),
		*cg = (char *)gd, *work;
	int	mask, i, j, m, m1, m2, len = sizeof(Gene)*GeneLen;
	if (len > 32) {
fprintf(stderr,"You must extend 'crossover()' for larger gene %d.\n",len);
		exit(1);
	}
	switch (CrossMode) {
		case CrsUniform:
		mask = lrand48();
		for (j = 0, m = 1; j < len; j ++, m <<= 1)
			cg[j] = (m & mask)? cp1[j] : cp2[j];
		break;
		case CrsOnePoint:
		if (lrand48() & 1) { work = cp1; cp1 = cp2; cp2 = work; }
		m = (lrand48() % (len - 1)) + 1;
		memcpy(cg,cp1,m);
		memcpy(cg+m,cp2+m,len-m);
		break;
		case CrsTwoPoints:
		if (lrand48() & 1) { work = cp1; cp1 = cp2; cp2 = work; }
		m1 = lrand48() % len;
		do m2 = lrand48() % len; while (m2 == m1);
		if (m1 > m2) { m = m1; m1 = m2; m2 = m; }
		if (m1 > 0) memcpy(cg,cp1,m1);
		memcpy(cg+m1,cp2+m1,m2-m1);
		memcpy(cg+m2,cp1+m2,len-m2);
	}
}
mutation(gs,gd)
Gene	*gs, *gd;
{
	int	i, k;
	unsigned char	*p = (unsigned char *)gd;
	double	drand48();
	if (gs != gd) memcpy(gd,gs,GeneLen*sizeof(Gene));
	for (i = 0; i < GeneLen*sizeof(Gene); i ++, p ++)
		for (k = 0x80; k; k >>= 1)
			if (drand48() < MutationRate) *p ^= k;
}
